'''
## This file is autogenerated from macrosynergy setup.py.

It contains version information for the macrosynergy package.

short_version: 1.0.2

version: 1.0.2

full_version: 1.0.2

git_revision: 949600e6def935b323c822b221c3258420aee441

release: True

'''
short_version: str = '1.0.2'
version: str = '1.0.2'
full_version: str = '1.0.2'
git_revision: str = '949600e6def935b323c822b221c3258420aee441'
release: bool = True
if not release:
    version = full_version
